# -*- coding: utf-8 -*-
from django.test import TestCase
from .common import CommonResultTestCase
from ..models import Result, Seat, TeamSeason, ContenderSeason

class ResultTestCase(TestCase, CommonResultTestCase):
    def test_result_shorcuts(self):
        result = self.get_test_result()
        self.assertEquals(result.driver, result.seat.contender.driver)
        self.assertEquals(result.team, result.seat.team)

    def test_result_fastest_unique(self):
        result = self.get_test_result()
        result.fastest_lap = True
        self.assertIsNone(result.save())
        seat_a = result.seat
        race = result.race
        self.assertEquals(Result.objects.get(race=race, seat=seat_a).fastest_lap, True)
        # create result_b
        seat_b = self.get_test_seat_b(seat_a=seat_a)
        result_b = self.get_test_result(seat=seat_b, race=race)
        result_b.fastest_lap = True
        self.assertIsNone(result_b.save())
        # Although both results have fastest_lap=True, only the last is saved.
        self.assertEquals(Result.objects.filter(race=race, fastest_lap=True).count(), 1)
        self.assertEquals(Result.objects.get(race=race, seat=seat_a).fastest_lap, False)
        self.assertEquals(Result.objects.get(race=race, seat=seat_b).fastest_lap, True)

    def test_result_points(self):
        result = self.get_test_result()
        seat = result.seat
        race = result.race
        season = race.season
        # season.punctuation = 'F1-25'
        # self.assertIsNone(season.save())

        result.qualifying = 2
        result.finish = 2
        self.assertIsNone(result.save())

        # result.finish = 2 = ?? points
        self.assertGreater(result.points, 0)
        team_season = TeamSeason.objects.get(team=seat.team, season=season)
        self.assertGreater(team_season.get_points(), 0)
        race_points = result.points

        # change scoring to add fastest_lap point
        result.fastest_lap = True
        self.assertIsNone(result.save())

        # modify scoring to check if fastest_lap scoring is counted
        scoring = season.get_scoring()
        scoring['fastest_lap'] = 1
        with self.settings(DR27_CONFIG={'PUNCTUATION': {'F1-25': scoring}}):
            # result.points is greater than before
            self.assertGreater(result.points, race_points)  # No working currently
        self.assertEquals(race.fastest, result.seat)

    def test_result_seat_exception(self):
        self.get_test_result(raise_seat_exception=True)

    def test_result_team_exception(self):
        self.get_test_result(raise_team_exception=True)

    def test_rank(self):
        result_a = self.get_test_result(qualifying=2, finish=1)
        seat_a = result_a.seat
        race = result_a.race
        # set season
        season = race.season
        # season.punctuation = 'F1-25'
        # self.assertIsNone(season.save())

        seat_b = self.get_test_seat_b(seat_a=seat_a)
        result_b = self.get_test_result(seat=seat_b, race=race, qualifying=1, finish=3)

        # pole
        self.assertEquals(race.pole, seat_b)
        # winner
        self.assertEquals(race.winner, seat_a)

        # season contender = 2
        self.assertEquals(len(season.contenders()), 2)
        self.assertEquals(len(season.points_rank()), 2)
        self.assertEquals(len(season.olympic_rank()), 2)
        self.assertEquals(len(season.team_points_rank()), 1) # seat a and seat b is in the same team
        self.assertEquals(season.leader[1], seat_a.contender.driver) # seat a.driver is the winner, the leader
        self.assertEquals(season.team_leader[1], seat_a.team)

        # contenderseason.get_points
        contender_a = seat_a.contender
        contender_season = ContenderSeason(contender=contender_a, season=season)
        self.assertGreater(contender_season.get_points(), 0)
        self.assertGreater(contender_season.get_points(1), 0)



# -*- coding: utf-8 -*-
import sys

from django.core.exceptions import ValidationError
from ..models import Driver, Competition, Team, Season, Circuit
from ..models import Contender, TeamSeason, Seat, GrandPrix, Race, Result, ContenderSeason

def retro_encode(text):
    if sys.version_info < (3, 0):
        try:
            return text.encode('utf-8')
        except UnicodeDecodeError:
            return text
    else:
        return text

class CommonCreateTestCase(object):
    def set_test_create(self, model, **kwargs):
        self.assertTrue(model.objects.create(**kwargs))
        return model.objects.get(**kwargs)


class CommonDriverTestCase(object):
    def set_test_driver(self, **kwargs):
        self.assertTrue(Driver.objects.create(year_of_birth=1985, country='ES', **kwargs))
        return Driver.objects.get(**kwargs)

    def get_test_driver_a(self):
        test_driver_args = {'last_name': 'García', 'first_name': 'Juan'}
        return self.set_test_driver(**test_driver_args)

    def get_test_driver_b(self):
        test_driver_args = {'last_name': 'López', 'first_name': 'Jaime'}
        return self.set_test_driver(**test_driver_args)


class CommonCompetitionTestCase(CommonCreateTestCase):
    def set_test_competition(self, **kwargs):
        return self.set_test_create(model=Competition, **kwargs)

    def get_test_competition_a(self):
        test_competition_args = {'name': 'Competición A', 'full_name': 'Competición ABC'}
        return self.set_test_competition(**test_competition_args)

    def get_test_competition_b(self):
        test_competition_args = {'name': 'Competición B', 'full_name': 'Competición BDF'}
        return self.set_test_competition(**test_competition_args)

class CommonTeamTestCase(CommonCreateTestCase):
    def set_test_team(self, **kwargs):
        return self.set_test_create(model=Team, **kwargs)

    def get_test_team(self):
        team_args = {'name': 'Escudería Tec Auto', 'full_name': 'Escudería Tec Auto', 'country': 'ES'}
        return self.set_test_team(**team_args)

    def get_test_team_b(self):
        team_args = {'name': 'D27R', 'full_name': 'Driver 27 Racing Team', 'country': 'ES'}
        return self.set_test_team(**team_args)

class CommonContenderTestCase(CommonDriverTestCase, CommonCompetitionTestCase):
    def set_test_contender(self, driver, competition):
        contender_args = {'driver': driver, 'competition': competition}
        self.assertTrue(Contender.objects.create(**contender_args))
        return Contender.objects.get(**contender_args)

    def get_test_contender(self):
        driver = self.get_test_driver_a()
        competition = self.get_test_competition_a()
        return self.set_test_contender(driver=driver, competition=competition)

    def get_test_contender_b(self, contender):
        driver = self.get_test_driver_b()
        competition = contender.competition
        return self.set_test_contender(driver=driver, competition=competition)



class CommonSeasonTestCase(CommonCompetitionTestCase):
    def set_test_season(self, **kwargs):
        return self.set_test_create(model=Season, **kwargs)

    def get_test_season(self, competition):
        season_args = {'year': '2016', 'competition': competition, 'punctuation': 'F1-25', 'rounds': 10}
        return self.set_test_season(**season_args)

class CommonTeamSeasonTestCase(CommonTeamTestCase, CommonSeasonTestCase):
    pass

class CommonSeatTestCase(CommonContenderTestCase, CommonTeamTestCase):
    def set_test_seat(self, **kwargs):
        self.assertTrue(Seat.objects.create(**kwargs))
        seat = Seat.objects.get(**kwargs)
        return seat

    def get_test_seat(self):
        contender = self.get_test_contender()
        competition = contender.competition
        team = self.get_test_team()
        seat_args = {'contender': contender, 'team': team}
        # Team 1 not is a team of Competition A
        self.assertRaises(ValidationError, Seat.objects.create, **seat_args)
        # Add Team 1 to Competition A
        self.assertIsNone(team.competitions.add(competition))
        # Seat 1 OK
        return self.set_test_seat(**seat_args)

    def get_test_seat_b(self, seat_a):
        # different contender, same team
        contender = self.get_test_contender_b(contender=seat_a.contender)
        team = seat_a.team
        seat_args = {'contender': contender, 'team': team}
        # Seat 2 OK
        return self.set_test_seat(**seat_args)

    def get_test_seat_c(self, seat_a):
        # same contender A, different team
        contender = seat_a.contender
        team = self.get_test_team_b()
        # Add Team 1 to Competition A
        self.assertIsNone(team.competitions.add(contender.competition))
        seat_args = {'contender': contender, 'team': team}
        # Seat 3 OK
        return self.set_test_seat(**seat_args)


class CommonRaceTestCase(CommonSeasonTestCase):
    def get_test_circuit(self):
        circuit_args = {"name": "Autódromo de Jacarepaguá", "city": "Jacarepaguá"}
        self.assertTrue(Circuit.objects.create(country='BR', opened_in=1978, **circuit_args))
        circuit = Circuit.objects.get(**circuit_args)
        return circuit

    def get_test_grandprix(self):
        circuit = self.get_test_circuit()
        grandprix_args = {'name': 'Grande Prêmio do Brasil'}
        self.assertTrue(GrandPrix.objects.create(
            default_circuit=circuit, country='BR', first_held=1972, **grandprix_args))
        return GrandPrix.objects.get(**grandprix_args)

    def get_test_race(self, competition = None, round=1):
        if not competition:
            competition = self.get_test_competition_a()
        season = self.get_test_season(competition=competition)
        race_args = {
            'round': round,
            'season': season,
            'date': None,
            'alter_punctuation': None
        }
        # race without grandprix
        self.assertTrue(Race.objects.create(**race_args))
        race = Race.objects.get(season=season, round=1)
        return race

class CommonResultTestCase(CommonSeatTestCase, CommonRaceTestCase):
    def set_test_result(self, *args, **kwargs):
        self.assertTrue(Result.objects.create(**kwargs))
        result = Result.objects.get(**{'seat': kwargs['seat'], 'race': kwargs['race']})
        return result

    def get_result_seat(self, seat):
        if not seat:
            seat = self.get_test_seat()
        return seat

    def get_result_race(self, race, competition):
        if not race:
            race = self.get_test_race(competition=competition)
        return race


    def get_test_result(self, seat=None, race=None, qualifying=None, finish=None,
                        raise_seat_exception=False, raise_team_exception=False):
        # when raise_team_exception, the team-season will not be created
        # raise a exception, because the team is invalid in that season
        seat = self.get_result_seat(seat) # if seat is None, get default seat
        competition = seat.contender.competition
        race = self.get_result_race(race=race, competition=competition) # If Race is None, get default race
        result_args = {
            'seat': seat,
            'race': race,
            'qualifying': qualifying,
            'finish': finish,
            'fastest_lap': False,
            'retired': False,
            'wildcard': False,
            'comment': None
        }
        season = race.season
        if not raise_team_exception:
            if seat.team not in season.teams.all():
                self.assertTrue(TeamSeason.objects.create(team=seat.team, season=season))
            if not raise_seat_exception and seat not in season.seats.all():
                self.assertIsNone(seat.seasons.add(season))

        if raise_seat_exception or raise_team_exception:
            self.assertRaises(ValidationError, Result.objects.create, **result_args)
        else:
            return self.set_test_result(**result_args)


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, IntegrityError
from django.dispatch import receiver
from django.db.models.signals import m2m_changed, pre_save, pre_delete
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext as _
from .punctuation import get_punctuation_config
from slugify import slugify
from django_countries.fields import CountryField
from exclusivebooleanfield.fields import ExclusiveBooleanField


@python_2_unicode_compatible
class Driver(models.Model):
    """ Main Driver Model. To combine with competitions (Contender) and competition/team (Seat) """
    last_name = models.CharField(max_length=50, verbose_name=_('last name'))
    first_name = models.CharField(max_length=25, verbose_name=_('first name'))
    year_of_birth = models.IntegerField(verbose_name=_('year of birth'))
    country = CountryField(verbose_name=_('country'))
    competitions = models.ManyToManyField('Competition', through='Contender', related_name='drivers',
                                          verbose_name=_('competitions'))

    def clean(self):
        if self.year_of_birth < 1900 or self.year_of_birth > 2099:
            raise ValidationError(_('Year_of_birth must be between 1900 and 2099'))
        super(Driver, self).clean()

    def __str__(self):
        return ', '.join((self.last_name, self.first_name))

    class Meta:
        unique_together = ('last_name', 'first_name')
        ordering = ['last_name', 'first_name']
        verbose_name = _('Driver')
        verbose_name_plural = _('Drivers')


@python_2_unicode_compatible
class Competition(models.Model):
    """ Competition model. Season, races, results... are grouped by competition """
    name = models.CharField(max_length=30, verbose_name=_('competition'), unique=True)
    full_name = models.CharField(max_length=100, unique=True, verbose_name=_('full name'))
    country = CountryField(null=True, blank=True, default=None, verbose_name=_('country'))
    slug = models.SlugField(null=True, blank=True, default=None)

    def save(self, *args, **kwargs):
        self.slug = slugify(self.name)
        super(Competition, self).save(*args, **kwargs)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']
        verbose_name = _('Competition')
        verbose_name_plural = _('Competitions')


@python_2_unicode_compatible
class Contender(models.Model):
    """ Contender are used to ranks or individual stats"""
    " Contender only can be related with teams of same competition "
    " If driver is related with same team in different competition, it is necessary to create new contender "
    driver = models.ForeignKey(Driver, related_name='career', verbose_name=_('driver'))
    competition = models.ForeignKey('Competition', related_name='contenders', verbose_name=_('competition'))
    teams = models.ManyToManyField('Team', through='Seat', related_name='contenders',
                                   verbose_name=_('teams'))

    def get_season(self, season):
        try:
            contender_season = ContenderSeason(self, season)
            return contender_season
        except ValidationError:
            return None

    @property
    def teams_verbose(self):
        teams = self.teams
        return ', '.join([team.name for team in teams.all()]) if teams.count() else None

    def __str__(self):
        return _('%(driver)s in %(competition)s') % {'driver': self.driver, 'competition': self.competition}

    class Meta:
        unique_together = ('driver', 'competition')
        ordering = ['competition__name', 'driver__last_name', 'driver__first_name']
        verbose_name = _('Contender')
        verbose_name_plural = _('Contenders')


@python_2_unicode_compatible
class Team(models.Model):
    """ Team model, unique if is the same in different competition """
    name = models.CharField(max_length=75, verbose_name=_('team'), unique=True)
    full_name = models.CharField(max_length=200, unique=True, verbose_name=_('full name'))
    competitions = models.ManyToManyField('Competition', related_name='teams', verbose_name=_('competitions'))
    country = CountryField(verbose_name=_('country'))

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']
        verbose_name = _('Team')
        verbose_name_plural = _('Teams')


@python_2_unicode_compatible
class Seat(models.Model):
    """ Seat is contender/team relation """
    " If a same contender drives to two teams in same season/competition, he will have many seats as driven teams "
    " e.g. Max Verstappen in 2016 (Seat 1: Toro Rosso, Seat 2: Red Bull) "
    team = models.ForeignKey('Team', related_name='seats', verbose_name=_('team'))
    contender = models.ForeignKey('Contender', related_name='seats', verbose_name=_('contender'))
    current = ExclusiveBooleanField(on='contender', default=False, verbose_name=_('current team'))
    seasons = models.ManyToManyField('Season', related_name='seats', blank=True, default=None,
                                     verbose_name=_('seasons'), through='SeatSeason')

    def clean(self):
        if self.contender.competition not in self.team.competitions.all():
            raise ValidationError(
                _("%(team)s is not a team of %(competition)s") %
                {'team': self.team, 'competition': self.contender.competition}
            )
        super(Seat, self).clean()

    def __str__(self):
        return _('%(driver)s in %(team)s/%(competition)s') \
               % {'driver': self.contender.driver, 'team': self.team,
                  'competition': self.contender.competition}

    class Meta:
        unique_together = ('team', 'contender')
        ordering = ['current', 'contender__driver__last_name', 'team']
        verbose_name = _('Seat')
        verbose_name_plural = _('Seats')


@python_2_unicode_compatible
class Circuit(models.Model):
    """ Circuit model. It can be in any competition, grand_prix or season """
    name = models.CharField(max_length=30, verbose_name=_('circuit'), unique=True)
    city = models.CharField(max_length=100, null=True, blank=True, verbose_name=_('city'))
    country = CountryField(verbose_name=_('country'))
    opened_in = models.IntegerField(verbose_name=_('opened in'))

    # @todo Add Clockwise and length
    def __str__(self):
        # @todo Add country name in __str__
        return self.name

    class Meta:
        ordering = ['name']
        verbose_name = _('Circuit')
        verbose_name_plural = _('Circuits')


@python_2_unicode_compatible
class GrandPrix(models.Model):
    """ Grand Prix model """
    " default_circuit will be in future the default choice in circuit selector "
    name = models.CharField(max_length=30, verbose_name=_('grand prix'), unique=True)
    country = CountryField(null=True, blank=True, default=None, verbose_name=_('country'))
    first_held = models.IntegerField(null=True, blank=True, verbose_name=_('first held'))
    default_circuit = models.ForeignKey(Circuit, related_name='default_to_grands_prix', null=True,
                                        blank=True, default=None, verbose_name=_('default circuit'))
    competitions = models.ManyToManyField('Competition', related_name='grands_prix', default=None,
                                          verbose_name=_('competitions'))

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']
        verbose_name = _('Grand Prix')
        verbose_name_plural = _('Grands Prix')


@python_2_unicode_compatible
class Season(models.Model):
    """ Season model. The main model to restrict races, results and punctuation """
    year = models.IntegerField(verbose_name=_('year'))
    competition = models.ForeignKey(Competition, related_name='seasons', verbose_name=_('competition'))
    rounds = models.IntegerField(blank=True, null=True, default=None, verbose_name=_('rounds'))
    teams = models.ManyToManyField(Team, related_name='seasons', through='TeamSeason',
                                   verbose_name=_('teams'))
    punctuation = models.CharField(max_length=20, null=True, default=None, verbose_name=_('punctuation'))

    def get_scoring(self, code=None):
        """ Getting the punctuation config. Chosen punctuation will be override temporarily by code kwarg """
        if not code:
            code = self.punctuation
        return get_punctuation_config(code)

    def pending_races(self):
        """ Based on rounds field, return races count when race doesn't have any result """
        past_races = Race.objects.filter(season=self, results__pk__isnull=False).distinct().count()
        pending_races = (self.rounds - past_races)
        return pending_races

    def pending_points(self):
        """ Return the maximum of available points taking into account the number of pending races"""
        # @todo fastest_lap and pole points are not counted
        scoring = self.get_scoring()
        max_score_by_race = sorted(scoring['finish'], reverse=True)[0]
        pending_races = self.pending_races()
        return pending_races * max_score_by_race

    def has_champion(self):
        """ If pending points is less than different between leader and runner-up, there is a champions """
        leader_points = self.leader[0] if self.leader else 0
        runner_up_points = self.runner_up[0] if self.runner_up else 0
        pending_points = self.pending_points()
        leader_distance = leader_points - runner_up_points
        return leader_distance >= pending_points

    def contenders(self):
        """ As season is related with seats, this method is a shorcut to get contenders """
        seats = [seat.pk for seat in self.seats.all()]
        return Contender.objects.filter(seats__pk__in=seats).distinct()

    def stats_rank(self, **filters):
        """ Get driver rank based on record filter """
        contenders = self.contenders()
        rank = []
        for contender in contenders:
            contender_season = contender.get_season(self)
            rank.append((contender_season.get_stats(**filters), contender.driver, contender_season.teams_verbose))
        rank = sorted(rank, key=lambda x: x[0], reverse=True)
        return rank

    @staticmethod
    def get_team_rank_method(rank_type):
        """ Dict with distinct teams rank method, depending of rank_type param """
        " STATS: Count 1 by each time that driver get record "
        " RACES: Count 1 by each race that any driver get record "
        " DOUBLES: Count 1 by each race that at least two drivers get record "
        rank_dict = {
            'STATS': 'team_stats_rank',
            'RACES': 'team_races_rank',
            'DOUBLES': 'team_doubles_rank'
        }

        return rank_dict.get(rank_type)

    def get_team_rank(self, rank_type, **filters):
        """ Return team rank calling returned method by get_team_rank_method """
        rank_method = self.get_team_rank_method(rank_type)
        return getattr(self, rank_method)(**filters) if rank_method else None

    def team_rank(self, total_method, **filters):
        """ Collect the records of each team calling the method of TeamSeason passed by total_method param """
        rank = []
        teams = self.teams.all()
        for team in teams:
            team_season = TeamSeason.objects.get(season=self, team=team)
            total = getattr(team_season, total_method)(**filters)
            rank.append((total, team))
            rank = sorted(rank, key=lambda x: x[0], reverse=True)
        return rank

    def team_stats_rank(self, **filters):
        return self.team_rank('get_total_stats', **filters)

    def team_races_rank(self, **filters):
        return self.team_rank('get_total_races', **filters)

    def team_doubles_rank(self, **filters):
        return self.team_rank('get_doubles_races', **filters)

    def points_rank(self, scoring_code=None):
        """ Points driver rank. Scoring can be override by scoring_code param """
        contenders = self.contenders()
        scoring = self.get_scoring(scoring_code)
        rank = []
        for contender in contenders:
            contender_season = contender.get_season(self)
            rank.append((contender_season.get_points(scoring=scoring), contender.driver, contender_season.teams_verbose,
                         contender_season.get_positions_str()))
        rank = sorted(rank, key=lambda x: (x[0], x[3]), reverse=True)
        return rank

    def olympic_rank(self):
        """ The driver
        with superior race results (based on descending order, from number of
        wins to numbers of second-places down) will gain precedence. """
        contenders = self.contenders()
        rank = []
        for contender in contenders:
            contender_season = contender.get_season(self)
            position_list = contender_season.get_positions_list()
            position_str = contender_season.get_positions_str(position_list=position_list)
            rank.append((position_str,
                         contender.driver,
                         contender_season.teams_verbose,
                         position_list))
        rank = sorted(rank, key=lambda x: x[0], reverse=True)
        return rank

    def team_points_rank(self):
        """ Same that points_rank by count both team drivers """
        teams = self.teams.all()
        rank = []
        for team in teams:
            team_season = TeamSeason.objects.get(season=self, team=team)
            rank.append((team_season.get_points(), team))
        rank = sorted(rank, key=lambda x: x[0], reverse=True)
        return rank

    def get_leader(self, team=False):
        """ Get driver leader or team leader """
        if team:
            rank = self.team_points_rank()
        else:
            rank = self.points_rank()
        return rank[0] if len(rank) else None

    @property
    def leader(self):
        """ get_leader(driver) property """
        return self.get_leader()

    @property
    def runner_up(self):
        """ Get the runner_up (driver) """
        rank = self.points_rank()
        return rank[1] if len(rank) > 1 else None

    @property
    def team_leader(self):
        """ get_leader(team) property """
        return self.get_leader(team=True)

    def __str__(self):
        return '/'.join((self.competition.name, str(self.year)))

    class Meta:
        unique_together = ('year', 'competition')
        ordering = ['competition__name', 'year']
        verbose_name = _('Season')
        verbose_name_plural = _('Seasons')


class SeatsSeason(Season):
    """ Aux proxy model to use in Admin for Season/Seat relation from Season """
    class Meta:
        proxy = True
        verbose_name = _('Seats by season')
        verbose_name_plural = _('Seats by season')


@python_2_unicode_compatible
class Race(models.Model):
    """ Race model. It can altered season punctuation (double or half) only in this race"""
    ALTER_PUNCTUATION = (
        ('double', _('Double')),
        ('half', _('Half'))
    )
    season = models.ForeignKey(Season, related_name='races', verbose_name=_('season'))
    round = models.IntegerField(verbose_name=_('round'))
    grand_prix = models.ForeignKey(GrandPrix, related_name='races', blank=True, null=True,
                                   default=None, verbose_name=_('grand prix'))
    circuit = models.ForeignKey(Circuit, related_name='races', blank=True, null=True,
                                default=None, verbose_name=_('circuit'))
    date = models.DateField(blank=True, null=True, default=None, verbose_name=_('date'))
    alter_punctuation = models.CharField(choices=ALTER_PUNCTUATION, null=True, blank=True,
                                         default=None, max_length=6, verbose_name=_('alter punctuation'))

    def _grandprix_exception(self):
        """ Grand Prix must be related with season competition """
        competition = self.season.competition
        if competition and self.grand_prix and competition not in self.grand_prix.competitions.all():
            return True
        else:
            return False

    def clean(self, *args, **kwargs):
        """ Validate round and grand_prix field """
        errors = {}
        season = getattr(self, 'season', None)
        if season:
            if self.round > self.season.rounds:
                errors['round'] = _('Max rounds in this season: %(rounds)d') % {'rounds': self.season.rounds}
            if self._grandprix_exception():
                errors['grand_prix'] = _("%(grand_prix)s is not a/an %(competition)s Grand Prix") \
                                       % {'grand_prix': self.grand_prix, 'competition': self.season.competition}
        if errors:
            raise ValidationError(errors)
        super(Race, self).clean()

    def get_result_seat(self, **kwargs):
        """ Return the first result that match with filter """
        results = self.results.filter(**kwargs)
        return results.first().seat if results.count() else None

    @property
    def pole(self):
        """ get_result_seat(pole) """
        return self.get_result_seat(qualifying=1)

    @property
    def winner(self):
        """ get_result_seat(winner) """
        return self.get_result_seat(finish=1)

    @property
    def fastest(self):
        """ get_result_seat(fastest) """
        return self.get_result_seat(fastest_lap=True)

    def __str__(self):
        race_str = '{season}-{round}'.format(season=self.season, round=self.round)
        if self.grand_prix:
            race_str += u'.{grand_prix}'.format(grand_prix=self.grand_prix)
        return race_str

    class Meta:
        unique_together = ('season', 'round')
        ordering = ['season', 'round']
        verbose_name = _('Race')
        verbose_name_plural = _('Races')


class SeatSeason(models.Model):
    """ Model created to validate restriction between both models """
    seat = models.ForeignKey('Seat', related_name='seasons_seat')
    season = models.ForeignKey('Season', related_name='seats_season')

    @staticmethod
    def get_seat_season_errors(seat, season):
        """ Check if seat is valid in current season competition """
        seat_competition = seat.contender.competition
        season_competition = season.competition
        errors = []
        if seat_competition != season_competition:
            errors.append(
                _('%(season)s is not a/an %(competition)s season')
                % {'season': season, 'competition': seat_competition}
            )
        return errors

    @staticmethod
    def get_seat_team_season_error(team, season):
        """ Check if team is valid in current season """
        " Team must be related with season before of save Seat/Season rel "
        errors = []
        if season.pk:
            season_teams = [season_team.pk for season_team in season.teams.all()]
            if team.pk not in season_teams:
                errors.append(
                    _('%(team)s is not a team of %(season)s') % {'team': team, 'season': season}
                )
        return errors

    def clean(self, *args, **kwargs):
        """ Check validation """
        seat = self.seat
        season = self.season
        errors = []
        errors.extend(self.get_seat_season_errors(seat, season))
        errors.extend(self.get_seat_team_season_error(seat.team, season))
        if errors:
            raise ValidationError(errors)
        super(SeatSeason, self).clean()

    def save(self, *args, **kwargs):
        """ Force to validate before of save """
        self.clean()
        super(SeatSeason, self).save(*args, **kwargs)

    class Meta:
        auto_created = True
        db_table = 'driver27_seat_seasons'


def seat_seasons(sender, instance, action, pk_set, **kwargs):  #noqa
    """ Signal in DriverCompetitionTeam.seasons to avoid seasons which not is in competition """
    if action == 'pre_add':
        errors = []
        for pk in list(pk_set):
            season = Season.objects.get(pk=pk)
            errors.extend(SeatSeason.get_seat_season_errors(instance, season))
            errors.extend(SeatSeason.get_seat_team_season_error(instance.team, season))
        if errors:
            raise ValidationError(errors)

m2m_changed.connect(seat_seasons, sender=SeatSeason)


@python_2_unicode_compatible
class TeamSeason(models.Model):
    """ TeamSeason model, only for validation although it is saved in DB"""
    season = models.ForeignKey('Season', related_name='teams_season', verbose_name=_('season'))
    team = models.ForeignKey('Team', related_name='seasons_team', verbose_name=_('team'))
    sponsor_name = models.CharField(max_length=75, null=True, blank=True, default=None,
                                    verbose_name=_('sponsor name'))

    def clean(self, *args, **kwargs):
        """ Check if team participate in competition """
        if hasattr(self, 'team') and hasattr(self, 'season'):
            team = self.team
            team_competitions = [competition.pk for competition in team.competitions.all()]
            if self.season.competition.pk not in team_competitions:
                raise ValidationError(
                    _('Team %(team)s doesn\'t participate in %(competition)s')
                    % {'team': self.team, 'competition': self.season.competition}
                )
        super(TeamSeason, self).clean()

    @staticmethod
    def delete_seat_exception(team, season):
        """ Force IntegrityError when delete a team with seats in season """
        if TeamSeason.check_delete_seat_restriction(team=team, season=season):
            raise IntegrityError('You cannot delete a team with seats in this season.'
                                 'Delete seats before')

    @staticmethod
    def check_delete_seat_restriction(team, season):
        seats_count = SeatSeason.objects.filter(seat__team=team, season=season).count()
        return bool(seats_count)
        # return {'team': _('Seats with %(team)s exists in this season. Delete seats before.' % {'team': team})}

    def get_results(self, limit_races=None):
        """ Return all results of team in season """
        results = Result.objects.filter(race__season=self.season, seat__team=self.team)
        if isinstance(limit_races, int):
            results = results.filter(race__round__lte=limit_races)
        results = results.order_by('race__round')
        return results

    def get_races(self, **filters):
        """ Return only race id of team in season """
        results = self.get_results().filter(**filters)\
            .values('race').annotate(count_race=models.Count('race'))\
            .order_by()
        return results

    def get_points(self):
        """ Get points. Exclude wildcards """
        results = self.get_results().exclude(wildcard=True).order_by('race__round')
        points_list = [result.points for result in results.all() if result.points is not None]
        return sum(points_list)

    def get_filtered_results(self, **filters):
        """ Filter results """
        results = self.get_results()
        return results.filter(**filters)

    def get_filtered_races(self, **filters):
        """ Filter races"""
        # @todo Check if filter in return line is necessary
        races = self.get_races()
        return races.filter(**filters)

    def get_total_races(self, **filters):
        """ Only count 1 by race with any driver in filter """
        return self.get_filtered_races(**filters).count()

    def get_doubles_races(self, **filters):
        """ Only count 1 by race with at least two drivers in filter """
        return self.get_filtered_races(**filters).filter(count_race__gte=2).count()

    def get_total_stats(self, **filters):  # noqa
        """ Count 1 by each result """
        return self.get_filtered_results(**filters).count()

    def __str__(self):
        str_team = self.team.name
        if self.sponsor_name:
            str_team = self.sponsor_name
        return '{team} in {season}'.format(team=str_team, season=self.season)

    class Meta:
        unique_together = ('season', 'team')
        verbose_name = _('Team Season')
        verbose_name_plural = _('Teams Season')


@python_2_unicode_compatible
class Result(models.Model):
    """ Result model """
    race = models.ForeignKey(Race, related_name='results', verbose_name=_('race'))
    seat = models.ForeignKey(Seat, related_name='results', verbose_name=_('seat'))
    qualifying = models.IntegerField(blank=True, null=True, default=None, verbose_name=_('qualifying'))
    finish = models.IntegerField(blank=True, null=True, default=None, verbose_name=_('finish'))
    fastest_lap = ExclusiveBooleanField(on='race', default=False, verbose_name=_('fastest lap'))
    retired = models.BooleanField(default=False, verbose_name=_('retired'))
    wildcard = models.BooleanField(default=False, verbose_name=_('wildcard'))
    comment = models.CharField(max_length=250, blank=True, null=True, default=None, verbose_name=_('comment'))

    def clean(self, *args, **kwargs):
        seat_errors = []
        if self.seat.team not in self.race.season.teams.all():
            seat_errors.append(_('Team is not in current season'))
        if self.seat not in self.race.season.seats.all():
            seat_errors.append(_('Seat is not in current season'))
        if seat_errors:
            raise ValidationError(_('Invalid Seat in this race. ')+'\n'.join(seat_errors))
        super(Result, self).clean()

    @property
    def driver(self):
        return self.seat.contender.driver

    @property
    def team(self):
        return self.seat.team

    def _get_fastest_lap_points(self, scoring):
        """ Return fastest_lap point if fastest lap is scoring and driver get fastest lap """
        if 'fastest_lap' in scoring and self.fastest_lap:
            return scoring['fastest_lap']
        else:
            return 0

    def _get_race_points(self, scoring):
        """ Return race points with factor depending of alter_punctuation field"""
        race = self.race
        points_factor = {'double': 2, 'half': 0.5}
        factor = points_factor[race.alter_punctuation] if race.alter_punctuation in points_factor else 1
        points_scoring = sorted(scoring['finish'], reverse=True)
        if self.finish:
            scoring_len = len(points_scoring)
            if not self.finish > scoring_len:
                return points_scoring[self.finish - 1] * factor
        return 0

    def points_calculator(self, scoring):
        """ Return points. Scoring can be the season scoring or scoring param"""
        points = 0
        if not scoring:
            scoring = self.race.season.get_scoring()
        points += self._get_fastest_lap_points(scoring)
        points += self._get_race_points(scoring)
        return points if points > 0 else None

    @property
    def points(self):
        """ Return points based on season scoring """
        scoring = self.race.season.get_scoring()
        return self.points_calculator(scoring)

    def __str__(self):
        string = '{seat} ({race})'.format(seat=self.seat, race=self.race)
        if self.finish:
            string += ' - {finish}º'.format(finish=self.finish)
        else:
            string += ' - ' + _('OUT')
        return string

    class Meta:
        unique_together = [('race', 'seat'), ('race', 'qualifying'), ('race', 'finish')]
        ordering = ['race__season', 'race__round', 'finish', 'qualifying']
        verbose_name = _('Result')
        verbose_name_plural = _('Results')


class ContenderSeason(object):
    """ ContenderSeason is not a model. Only for validation and ranks"""
    contender = None
    season = None
    teams = None
    contender_teams = None
    teams_verbose = None

    def __init__(self, contender, season):
        if not isinstance(contender, Contender) or not isinstance(season, Season):
            raise ValidationError(_('contender is not a Contender or/and season is not a Season'))
        self.contender = contender
        self.season = season
        self.seats = Seat.objects.filter(contender__pk=self.contender.pk, seasons__pk=self.season.pk)
        self.teams = Team.objects.filter(seats__in=self.seats)
        self.teams_verbose = ', '.join([team.name for team in self.teams])

    def get_results(self, limit_races=None):
        """ Return results. Can be limited."""
        results = Result.objects.filter(race__season=self.season, seat__contender=self.contender)
        if isinstance(limit_races, int):
            results = results.filter(race__round__lte=limit_races)
        results = results.order_by('race__round')
        return results

    def get_filtered_results(self, **filters):
        """ Apply filters to contender-season results """
        results = self.get_results()
        return results.filter(**filters)

    def get_stats(self, **filters):
        """ Count 1 by each result """
        return self.get_filtered_results(**filters).count()

    def get_points(self, limit_races=None, scoring=None):
        """ Get points. Can be limited. Scoring will be overwrite temporarily"""
        results = self.get_results(limit_races=limit_races)
        points_list = []
        for result in results.all():
            result_points = result.points_calculator(scoring)
            if result_points:
                points_list.append(result_points)
        return sum(points_list)
        
    def get_positions_list(self, limit_races=None):
        """ Return a list with the count of each 20 first positions """
        results = self.get_results(limit_races=limit_races)
        last_position = 20
        positions = []
        for x in range(1, last_position+1):
            position_count = results.filter(finish=x).count()
            positions.append(position_count)
        return positions

    def get_positions_str(self, position_list=None, limit_races=None):
        """ Return a str with position_list to order """
        " Each list item will be filled to zeros until get three digits e.g. 1 => 001, 12 => 012 "
        if not position_list:
            position_list = self.get_positions_list(limit_races=limit_races)
        positions_str = ''.join([str(x).zfill(3) for x in position_list])
        return positions_str


@receiver(pre_save)
def pre_save_handler(sender, instance, *args, **kwargs):
    """ Force full_clean in each model to validation """
    model_list = (Driver, Team, Competition, Contender, Seat, Circuit,
                  GrandPrix, Race, Season, Result, SeatSeason, TeamSeason)
    if not isinstance(instance, model_list):
        return
    instance.full_clean()


@receiver(pre_delete, sender=TeamSeason)
def pre_delete_team_season(sender, instance, *args, **kwargs):
    """ Force exception before of delete team in season when exists dependent seats """
    team = instance.team
    season = instance.season
    TeamSeason.delete_seat_exception(team=team, season=season)



from .config import get_config


def get_record_config(record_code=None):
    config = get_config('RECORDS', key=record_code)
    return config


def get_record_label_dict(doubles=False):
    config = get_record_config()
    if config:
        if doubles:
            return [(record_dict.get('label'), index) for index, record_dict in config.items()
                    if record_dict.get('doubles')]
        else:
            return [(record_dict.get('label'), index) for index, record_dict in config.items()]
    return None



# -*- coding: utf-8 -*-
from django.test import TestCase
from django.core.exceptions import ValidationError
from .common import CommonSeatTestCase, CommonSeasonTestCase
from ..models import ContenderSeason


class ContenderSeasonTestCase(TestCase, CommonSeatTestCase, CommonSeasonTestCase):
    def test_contender_season(self):
        contender = None
        season = None
        self.assertRaises(ValidationError, ContenderSeason,
                          **{'contender': contender, 'season': season})
        seat = self.get_test_seat()
        contender = seat.contender
        self.assertIsNone(contender.get_season(season))
        season = self.get_test_season(competition=contender.competition)
        self.assertTrue(ContenderSeason(contender=contender, season=season))
        self.assertIsInstance(contender.get_season(season), ContenderSeason)


import os
from .settings import *

# DATABASES = {
#     'default': {
#         'ENGINE': 'django.db.backends.sqlite3',
#         'NAME': os.path.join(BASE_DIR, 'pytest.sqlite3'),
#     }
# }

PYTEST_SETTING = True
ALLOWED_HOSTS.append('testserver')


from django.conf.urls import url, include

from driver27 import views

urlpatterns = [
    url(r'^$', views.competition_view, name='competition-list'),
    url(r'^(?P<competition_slug>[-\w\d]+)$', views.competition_view, name='competition-view'),
    url(r'^(?P<competition_slug>[-\w\d]+)/(?P<year>\d+)/', include('driver27.urls.season')),
]


from django.db.models.fields import BLANK_CHOICE_DASH
from .common import *
from ..punctuation import get_punctuation_label_dict
from ..models import Season


class SeasonAdminForm(AlwaysChangedModelForm):

    def __init__(self, *args, **kwargs):
        super(SeasonAdminForm, self).__init__(*args, **kwargs)
        punctuation_choices = get_punctuation_label_dict()
        self.fields['punctuation'] = forms.ChoiceField(choices=BLANK_CHOICE_DASH + list(punctuation_choices),
                                                       initial=None)

    class Meta:
        model = Season
        fields = ('year', 'competition', 'rounds', 'punctuation')



from django.conf.urls import url

from driver27 import views

urlpatterns = [
    url(r'^$', views.season_view, name='season-view'),
    url(r'^driver$', views.driver_rank_view, name='season-driver'),
    url(r'^driver/road-to-championship$', views.driver_road_view, name='season-driver-road'),
    url(r'^driver/olympic$', views.driver_olympic_view, name='season-driver-olympic'),
    url(r'^driver/record$', views.driver_record_view, name='season-driver-record-index'),
    url(r'^driver/record/(?P<record>[-\w\d]+)$', views.driver_record_view, name='season-driver-record'),
    url(r'^team$', views.team_rank_view, name='season-team'),
    url(r'^team/record$', views.team_record_stats_view, name='season-team-record-index'),
    url(r'^team/record/(?P<record>[-\w\d]+)$', views.team_record_stats_view, name='season-team-record'),
    url(r'^team/record/(?P<record>[-\w\d]+)/races$', views.team_record_races_view, name='season-team-record-races'),
    url(r'^team/record/(?P<record>[-\w\d]+)/doubles-in-race$', views.team_record_doubles_view,
        name='season-team-record-doubles'),
    url(r'^race$', views.race_list, name='season-race-list'),
    url(r'^race/(?P<race_id>\d+)$', views.race_view, name='season-race-view'),
    url(r'^race/(?P<race_id>\d+)/(?P<rank>[-\w\d]+)$', views.race_view, name='season-race-rank'),
]


# -*- coding: utf-8 -*-
from django.test import TestCase
from django.core.exceptions import ValidationError
from django.db import transaction, IntegrityError
from .common import CommonSeasonTestCase, CommonSeatTestCase
from ..models import SeatSeason, TeamSeason, Contender, Seat


class SeasonTestCase(TestCase, CommonSeasonTestCase, CommonSeatTestCase):
    def test_season_unicode(self):
        competition = self.get_test_competition_a()
        season = self.get_test_season(competition)
        year = season.year
        expected_season = '{competition}/{year}'.format(competition=competition, year=year)
        self.assertEquals(str(season), expected_season)

    def test_season_scoring(self):
        competition = self.get_test_competition_a()
        season = self.get_test_season(competition)
        self.assertIsInstance(season.get_scoring(), dict)
        season.punctuation = 'BLABLABLA'
        self.assertIsNone(season.save())
        self.assertIsNone(season.get_scoring())

    def test_season_has_champion(self):
        competition = self.get_test_competition_a()
        season = self.get_test_season(competition)
        self.assertFalse(season.has_champion())

    def test_season_pending_races(self):
        competition = self.get_test_competition_a()
        season = self.get_test_season(competition)
        self.assertIsInstance(season.pending_races(), int)

    def test_season_pending_points(self):
        competition = self.get_test_competition_a()
        season = self.get_test_season(competition)
        self.assertIsInstance(season.pending_points(), (int, float))

    def test_season_contenders(self):
        seat_a = self.get_test_seat()
        team_a = seat_a.team
        competition = seat_a.contender.competition
        season = self.get_test_season(competition)
        seat_b = self.get_test_seat_b(seat_a=seat_a)
        seat_c = self.get_test_seat_c(seat_a=seat_a)
        team_c = seat_c.team
        # create TeamSeason rel, A and B is the same
        self.assertTrue(TeamSeason.objects.create(team=team_a, season=season))
        self.assertTrue(TeamSeason.objects.create(team=team_c, season=season))
        # create SeatSeason rel
        self.assertIsNone(seat_a.seasons.add(season))
        self.assertIsNone(seat_b.seasons.add(season))
        self.assertIsNone(seat_c.seasons.add(season))
        # A and C is the same, 2 = (A, B)
        self.assertEquals(season.contenders().count(), 2)

    def test_seat_season_exception(self):
        seat_a = self.get_test_seat()
        competition = self.get_test_competition_b()
        season = self.get_test_season(competition)
        team = seat_a.team
        with transaction.atomic():
            self.assertRaises(ValidationError, seat_a.seasons.add, season)
        # generate new seat (same driver, new competition, same team)
        contender_b_args = {'competition': competition, 'driver': seat_a.contender.driver}
        self.assertTrue(self.set_test_contender(**contender_b_args))
        contender_b = Contender.objects.get(**contender_b_args)
        # add team to competition
        self.assertIsNone(team.competitions.add(competition))
        # save seat_b
        seat_b_args = {'contender': contender_b, 'team': team}
        self.assertTrue(self.set_test_seat(**seat_b_args))
        seat_b = Seat.objects.get(**seat_b_args)
        # Fail again because Seat team is not in season
        seat_season_args = {'seat': seat_b, 'season': season}
        with transaction.atomic():
            self.assertRaises(ValidationError, seat_b.seasons.add, season)
        self.assertRaises(ValidationError, SeatSeason.objects.create, **seat_season_args)
        # Save team-season rel and Seat-Season save is ok
        team_season_args = {'team': team, 'season': season}
        self.assertTrue(TeamSeason.objects.create(**team_season_args))
        self.assertTrue(SeatSeason.objects.create(**seat_season_args))
        team_season_filter = TeamSeason.objects.filter(**team_season_args)
        self.assertRaises(IntegrityError, team_season_filter.delete)



from django.conf import settings
from django.contrib.admin.sites import AdminSite
from django.test import TestCase, Client, RequestFactory

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse
from django.forms.models import inlineformset_factory, formset_factory
from ..models import Season, Driver, Team, Competition, Circuit, Result
from ..models import GrandPrix, Race, Contender, Seat, SeatSeason, TeamSeason
from ..admin import SeasonAdmin, SeasonAdminForm, DriverAdmin, SeatSeasonAdmin, TeamAdmin
from ..admin import CompetitionAdmin, CircuitAdmin, GrandPrixAdmin
from ..admin import RaceAdmin, ContenderAdmin, RelatedCompetitionAdmin
from ..admin import RaceInline, SeatInline, SeatSeasonInline, TeamSeasonInline
from ..admin.formsets import RaceFormSet, SeatSeasonFormSet, TeamSeasonFormSet
from ..admin.common import AlwaysChangedModelForm

class MockRequest(object):
    pass

class MockSuperUser(object):
    def has_perm(self, perm):
        return True

def get_request():
    request = MockRequest()
    request.user = MockSuperUser()
    return request


def get_fixtures_test():
    # django test uses a new db, while pytest use transaction no_commit in the same settings db
    if hasattr(settings, 'PYTEST_SETTING') and settings.PYTEST_SETTING:
        return None
    else:
        return ['circuits-2016.json', 'competition.json', 'drivers-2016.json', 'grands-prix.json',
                'seasons-2016.json', 'teams-2016.json', 'teams-season.json', 'driver-competition-2016.json',
                'driver-competition-team-2016.json',
                'races-2016.json', 'results-2016.json']


class FixturesTest(TestCase):
    fixtures = get_fixtures_test()


class ViewTest(FixturesTest):

    def setUp(self):
        self.site = AdminSite()
        self.client = Client()
        self.factory = RequestFactory()

    def test_competition_list(self):
        # Issue a GET request.
        response = self.client.get(reverse('competition-list'))

        # Check that the response is 302 OK.
        self.assertEqual(response.status_code, 200)

    def test_competition_view(self):
        # Issue a GET request.
        response = self.client.get(reverse('competition-view', kwargs={'competition_slug': 'f1'}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 200)

        response = self.client.get(reverse('competition-view', kwargs={'competition_slug': 'f19'}))
        # Check that the response is 404.
        self.assertEqual(response.status_code, 404)

    def test_season_view(self):
        # Issue a GET request.
        response = self.client.get(reverse('season-view', kwargs={'competition_slug': 'f1', 'year': 2016}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 200)

        response = self.client.get(reverse('season-driver', kwargs={'competition_slug': 'f1', 'year': 2016}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 200)

        response = self.client.get(reverse('season-driver-olympic', kwargs={'competition_slug': 'f1', 'year': 2016}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 200)

        response = self.client.get(reverse('season-driver-road', kwargs={'competition_slug': 'f1', 'year': 2016}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 200)

        response = self.client.get(reverse('season-team', kwargs={'competition_slug': 'f1', 'year': 2016}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 200)

        response = self.client.get(reverse('season-race-list', kwargs={'competition_slug': 'f1', 'year': 2016}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 200)

        response = self.client.get(reverse('season-view', kwargs={'competition_slug': 'f19', 'year': 2006}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 404)

    def test_race_view(self):
        # Issue a GET request.
        response = self.client.get(reverse('season-race-view', kwargs={'competition_slug': 'f1', 'year': 2016, 'race_id':1}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 200)
        # Issue a GET request.
        response = self.client.get(reverse('season-race-view', kwargs={'competition_slug': 'f1', 'year': 2016, 'race_id': 200}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 404)

    def test_driver_records_view(self):
        # Issue a GET request.
        response = self.client.get(reverse('season-driver-record-index',
                                           kwargs={'competition_slug': 'f1', 'year': 2016}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 200)

        # Issue a GET request.
        response = self.client.get(reverse('season-driver-record',
                                           kwargs={'competition_slug': 'f1', 'year': 2016, 'record': 'POLE'}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 200)

        # Issue a GET request.
        response = self.client.get(reverse('season-driver-record',
                                           kwargs={'competition_slug': 'f1', 'year': 2016, 'record': 'FFF'}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 404)

    def test_team_records_view(self):
        # Issue a GET request.
        response = self.client.get(reverse('season-team-record-index',
                                           kwargs={'competition_slug': 'f1', 'year': 2016}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 200)

        # Issue a GET request.
        response = self.client.get(reverse('season-team-record',
                                           kwargs={'competition_slug': 'f1', 'year': 2016, 'record': 'POLE'}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 200)

        # Issue a GET request.
        response = self.client.get(reverse('season-team-record-races',
                                           kwargs={'competition_slug': 'f1', 'year': 2016, 'record': 'POLE'}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 200)

        # Issue a GET request.
        response = self.client.get(reverse('season-team-record-doubles',
                                           kwargs={'competition_slug': 'f1', 'year': 2016, 'record': 'POLE'}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 200)

        # Issue a GET request.
        response = self.client.get(reverse('season-team-record',
                                           kwargs={'competition_slug': 'f1', 'year': 2016, 'record': 'FFF'}))
        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 404)

    # def test_admin(self):
    #     response = self.client.get(reverse('admin:driver27_driver_changelist'))
    #     # Check that the response is 200 OK.
    #     self.assertEqual(response.status_code, 302)
    #
    #     response = self.client.get(reverse('admin:driver27_driver_change', args=[1]))
    #     # Check that the response is 200 OK.
    #     self.assertEqual(response.status_code, 302)
    #
    #     response = self.client.get(reverse('admin:driver27_season_changelist'))
    #     # Check that the response is 200 OK.
    #     self.assertEqual(response.status_code, 302)
    #
    #     response = self.client.get(reverse('admin:driver27_season_change', args=[1]))
    #     # Check that the response is 200 OK.
    #     self.assertEqual(response.status_code, 302)
    #
    #     response = self.client.get(reverse('admin:driver27_race_changelist'))
    #     # Check that the response is 200 OK.
    #     self.assertEqual(response.status_code, 302)
    #
    #     response = self.client.get(reverse('admin:driver27_race_change', args=[1]))
    #     # Check that the response is 200 OK.
    #     self.assertEqual(response.status_code, 302)

    def _check_get_changelist(self, ma):
        request = get_request()
        self.assertTrue(ma.get_changelist(request=request))

    def test_season_admin(self):
        ma = SeasonAdmin(Season, self.site)
        self._check_get_changelist(ma)
        self.assertTrue(ma.get_form(request=None, obj=None))
        self.assertIsInstance(ma.get_season_copy(copy_id=1), dict)

        request = get_request()
        season = Season.objects.get(pk=1)
        season_form = ma.get_form(request=request, obj=season)
        self.assertTrue(ma.print_copy_season(obj=season))
        self.assertIsNotNone(SeasonAdminForm(season_form))
        request = self.factory.request(QUERY_STRING='copy=1')
        self.assertTrue(ma.get_changeform_initial_data(request=request))

    def test_driver_admin(self):
        ma = DriverAdmin(Driver, self.site)
        self._check_get_changelist(ma)
        driver = Driver.objects.get(pk=1)
        self.assertTrue(ma.print_competitions(driver))

    def test_team_admin(self):
        ma = TeamAdmin(Team, self.site)
        self._check_get_changelist(ma)
        team = Team.objects.get(pk=1)
        self.assertTrue(ma.print_competitions(team))

    def test_competition_admin(self):
        ma = CompetitionAdmin(Competition, self.site)
        self._check_get_changelist(ma)

    def test_circuit_admin(self):
        ma = CircuitAdmin(Circuit, self.site)
        self._check_get_changelist(ma)

    def test_grandprix_admin(self):
        ma = GrandPrixAdmin(GrandPrix, self.site)
        self._check_get_changelist(ma)

    def _test_season_formset(self, child_model, formset, fields):
        inline_formset = inlineformset_factory(Season, child_model, formset=formset,
                                               fields=fields,
                                               form=AlwaysChangedModelForm, can_delete=True)
        return inline_formset

    def _test_season_formset_copy(self, child_model, formset, fields, data=None):
        inline_formset = self._test_season_formset(child_model, formset, fields)
        inline_formset.request = self.factory.request(QUERY_STRING='copy=1')
        related_formset = inline_formset(data)
        self.assertTrue(related_formset.get_copy(copy_id=1))
        # self.assertFalse(related_formset.is_empty_form())
        self.assertFalse(related_formset.has_changed())
        return related_formset

    def test_race_formset(self):
        self._test_season_formset_copy(Race, RaceFormSet, ('round', 'grand_prix', 'circuit', 'date', 'alter_punctuation'))

    def test_seat_season_formset(self):
        self._test_season_formset_copy(SeatSeason, SeatSeasonFormSet, ('seat', 'season',))

    def test_team_season_delete(self):
        season_post = {
            'teams_season-TOTAL_FORMS': '1',
            'teams_season-INITIAL_FORMS': '0',
            'teams_season-MAX_NUM_FORMS': '',
            'teams_season-0-id': 1,
            'teams_season-0-team': 1,
            'teams_season-0-season': 1,
            'teams_season-0-DELETE': True
        }
        formset = self._test_season_formset(TeamSeason, formset=TeamSeasonFormSet, fields=('team', 'season',))
        formset.request = self.factory.request(QUERY_STRING='copy=1')
        team_season_formset = formset(season_post)
        # Invalid form. Team contains seats in that season.
        self.assertFalse(team_season_formset.is_valid())

    def test_race_admin(self):
        ma = RaceAdmin(Race, self.site)
        self._check_get_changelist(ma)
        race = Race.objects.get(pk=1)
        self.assertEquals(ma.print_pole(race), str(race.pole.contender.driver))
        self.assertEquals(ma.print_winner(race), str(race.winner.contender.driver))
        self.assertEquals(ma.print_fastest(race), str(race.fastest.contender.driver))
        self.assertEquals(ma.clean_position('1'), 1)
        self.assertIsNone(ma.clean_position(''))

    def test_race_inline(self):
        race = Race.objects.get(pk=1)
        race_ma = RaceInline(SeasonAdmin, self.site)
        request = get_request()
        # self.assertIsNotNone(race_ma.get_formset())
        season = race.season
        request._obj_ = season
        self.assertIsNotNone(race_ma.formfield_for_foreignkey(Race.grand_prix.field, request=request))

    def test_race_with_no_results(self):
        ma = RaceAdmin(Race, self.site)
        race = Race.objects.get(pk=20)  # No results
        self.assertIsNone(ma.print_pole(race))
        self.assertIsNone(ma.print_winner(race))
        self.assertIsNone(ma.print_fastest(race))

    def test_contender_admin(self):
        ma = ContenderAdmin(Contender, self.site)
        self._check_get_changelist(ma)
        contender = Contender.objects.get(pk=1)
        self.assertIsNotNone(ma.print_current(contender))
        request = get_request()
        self.assertTrue(ma.get_form(request=request, obj=contender))

    def _check_formfield_for_foreignkey(self, ma, request_obj, dbfield):
        request = get_request()
        request._obj_ = request_obj
        self.assertIsNotNone(ma.formfield_for_foreignkey(dbfield, request=request))

    def test_seat_inline_admin(self):
        ma = SeatInline(ContenderAdmin, self.site)
        contender = Contender.objects.get(pk=1)
        self._check_formfield_for_foreignkey(ma, request_obj=contender, dbfield=Seat.team.field)

    def test_seat_season_admin(self):
        ma = SeatSeasonAdmin(SeatSeason, self.site)
        seat_season = SeatSeason.objects.get(pk=1)
        request = get_request()
        self.assertFalse(ma.has_add_permission(request=request))
        self.assertTrue(ma.get_form(request=request, obj=seat_season))

    def test_seat_season_inline_admin(self):
        ma = SeatSeasonInline(SeasonAdmin, self.site)
        season = Season.objects.get(pk=1)
        self._check_formfield_for_foreignkey(ma, request_obj=season, dbfield=Seat.seasons.through.seat.field)

    def test_team_season_inline_admin(self):
        ma = TeamSeasonInline(SeasonAdmin, self.site)
        season = Season.objects.get(pk=1)
        self._check_formfield_for_foreignkey(ma, request_obj=season, dbfield=TeamSeason.team.field)

    def test_related_competition_admin(self):
        race = Race.objects.get(pk=1)
        related_competition = RelatedCompetitionAdmin()
        # maybe exception
        self.assertIsNone(related_competition.print_competitions(race))



from django import forms
from ..models import Race
from ..models import Seat, SeatSeason
from ..models import TeamSeason


class RelatedWithSeasonFormSet(forms.models.BaseInlineFormSet):
    model = None
    model_attributes = []

    def get_copy(self, copy_id):
        related_objects = self.model.objects.filter(season__pk=copy_id)
        initial = []
        for x in related_objects:
            initial_x = {}
            for xa in self.model_attributes:
                initial_x[xa] = getattr(x, xa, None)
            initial.append(initial_x)
        return initial

    def __init__(self, *args, **kwargs):
        super(RelatedWithSeasonFormSet, self).__init__(*args, **kwargs)
        request = self.request
        copy_id = request.GET.get('copy')
        if not self.initial and copy_id:
            self.initial = self.get_copy(copy_id)
            self.extra += len(self.initial)


class RaceFormSet(RelatedWithSeasonFormSet):
    model = Race
    model_attributes = ('round', 'circuit', 'grand_prix',)


class SeatSeasonFormSet(RelatedWithSeasonFormSet):
    model = SeatSeason
    model_attributes = ('seat',)


class TeamSeasonFormSet(RelatedWithSeasonFormSet):
    model = TeamSeason
    model_attributes = ('team', 'season', 'sponsor_name',)

    def team_season_check(self, form):
        delete_checked = False
        if self.is_marked_to_delete(form):
            team_season_obj = form.cleaned_data.get('id')
            # team_season_obj = self.model.objects.get(pk=team_season_id)
            team = team_season_obj.team
            season = team_season_obj.season
            seat_check = self.model.check_delete_seat_restriction(team=team, season=season)
            if seat_check:
                delete_checked = True
        return delete_checked

    def clean(self):
        super(TeamSeasonFormSet, self).clean()
        delete_checked = False

        for form in self.forms:
            invalid_team_season_delete = self.team_season_check(form)
            if invalid_team_season_delete:
                delete_checked = True

        if delete_checked:
            raise forms.ValidationError('You cannot delete a team with seats in this season.'
                                        'Delete seats before')

    @staticmethod
    def is_marked_to_delete(form):
        marked_to_delete = False
        try:
            if form.cleaned_data and form.cleaned_data.get('DELETE'):
                marked_to_delete = True
        except AttributeError:
            pass
        return marked_to_delete



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-10-14 14:49
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_countries.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Circuit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30, unique=True, verbose_name='circuit')),
                ('city', models.CharField(blank=True, max_length=100, null=True)),
                ('country', django_countries.fields.CountryField(max_length=2)),
                ('opened_in', models.IntegerField()),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Competition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30, unique=True, verbose_name='competition')),
                ('full_name', models.CharField(max_length=100, unique=True)),
                ('country', django_countries.fields.CountryField(blank=True, default=None, max_length=2, null=True)),
                ('slug', models.SlugField(blank=True, default=None, null=True)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Contender',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('competition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='contenders', to='driver27.Competition')),
            ],
            options={
                'ordering': ['competition__name', 'driver__last_name', 'driver__first_name'],
            },
        ),
        migrations.CreateModel(
            name='Driver',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_name', models.CharField(max_length=50)),
                ('first_name', models.CharField(max_length=25)),
                ('year_of_birth', models.IntegerField()),
                ('country', django_countries.fields.CountryField(max_length=2)),
                ('competitions', models.ManyToManyField(related_name='drivers', through='driver27.Contender', to='driver27.Competition')),
            ],
            options={
                'ordering': ['last_name', 'first_name'],
            },
        ),
        migrations.CreateModel(
            name='GrandPrix',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30, unique=True, verbose_name='grand prix')),
                ('country', django_countries.fields.CountryField(blank=True, default=None, max_length=2, null=True)),
                ('first_held', models.IntegerField(blank=True, null=True)),
                ('competitions', models.ManyToManyField(default=None, related_name='grands_prix', to='driver27.Competition')),
                ('default_circuit', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='default_to_grands_prix', to='driver27.Circuit')),
            ],
            options={
                'ordering': ['name'],
                'verbose_name_plural': 'grands prix',
            },
        ),
        migrations.CreateModel(
            name='Race',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('round', models.IntegerField()),
                ('date', models.DateField(blank=True, default=None, null=True)),
                ('alter_punctuation', models.CharField(blank=True, choices=[('double', 'Double'), ('half', 'Half')], default=None, max_length=6, null=True)),
                ('circuit', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='races', to='driver27.Circuit')),
                ('grand_prix', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='races', to='driver27.GrandPrix')),
            ],
            options={
                'ordering': ['season', 'round'],
            },
        ),
        migrations.CreateModel(
            name='Result',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('qualifying', models.IntegerField(blank=True, default=None, null=True)),
                ('finish', models.IntegerField(blank=True, default=None, null=True)),
                ('fastest_lap', models.BooleanField(default=False)),
                ('retired', models.BooleanField(default=False)),
                ('wildcard', models.BooleanField(default=False)),
                ('comment', models.CharField(blank=True, default=None, max_length=250, null=True)),
                ('race', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='driver27.Race')),
            ],
            options={
                'ordering': ['race__season', 'race__round', 'finish', 'qualifying'],
            },
        ),
        migrations.CreateModel(
            name='Season',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('year', models.IntegerField()),
                ('rounds', models.IntegerField(blank=True, default=None, null=True)),
                ('punctuation', models.CharField(default=None, max_length=20, null=True)),
                ('competition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seasons', to='driver27.Competition')),
            ],
            options={
                'ordering': ['competition__name', 'year'],
            },
        ),
        migrations.CreateModel(
            name='Seat',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('current', models.BooleanField(default=False)),
                ('contender', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seats', to='driver27.Contender')),
                ('seasons', models.ManyToManyField(blank=True, default=None, related_name='seats', to='driver27.Season')),
            ],
            options={
                'ordering': ['contender__driver__last_name', 'team'],
            },
        ),
        migrations.CreateModel(
            name='Team',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=75, unique=True, verbose_name='team')),
                ('full_name', models.CharField(max_length=200, unique=True)),
                ('country', django_countries.fields.CountryField(max_length=2)),
                ('competitions', models.ManyToManyField(related_name='teams', to='driver27.Competition')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='TeamSeason',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sponsor_name', models.CharField(blank=True, default=None, max_length=75, null=True)),
                ('season', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='teams_season', to='driver27.Season')),
                ('team', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seasons_team', to='driver27.Team')),
            ],
            options={
                'verbose_name': 'Team Season',
                'verbose_name_plural': 'Teams Season',
            },
        ),
        migrations.AddField(
            model_name='seat',
            name='team',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seats', to='driver27.Team'),
        ),
        migrations.AddField(
            model_name='season',
            name='teams',
            field=models.ManyToManyField(related_name='seasons', through='driver27.TeamSeason', to='driver27.Team'),
        ),
        migrations.AddField(
            model_name='result',
            name='seat',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='driver27.Seat'),
        ),
        migrations.AddField(
            model_name='race',
            name='season',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='races', to='driver27.Season'),
        ),
        migrations.AddField(
            model_name='contender',
            name='driver',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='career', to='driver27.Driver'),
        ),
        migrations.AddField(
            model_name='contender',
            name='teams',
            field=models.ManyToManyField(related_name='contenders', through='driver27.Seat', to='driver27.Team'),
        ),
        migrations.AlterUniqueTogether(
            name='teamseason',
            unique_together=set([('season', 'team')]),
        ),
        migrations.AlterUniqueTogether(
            name='seat',
            unique_together=set([('team', 'contender')]),
        ),
        migrations.AlterUniqueTogether(
            name='season',
            unique_together=set([('year', 'competition')]),
        ),
        migrations.AlterUniqueTogether(
            name='result',
            unique_together=set([('race', 'seat'), ('race', 'qualifying'), ('race', 'finish')]),
        ),
        migrations.AlterUniqueTogether(
            name='race',
            unique_together=set([('season', 'round')]),
        ),
        migrations.AlterUniqueTogether(
            name='driver',
            unique_together=set([('last_name', 'first_name')]),
        ),
        migrations.AlterUniqueTogether(
            name='contender',
            unique_together=set([('driver', 'competition')]),
        ),
    ]



from django.conf.urls import url
from django.shortcuts import render
from django.utils.translation import ugettext as _
from .common import CommonTabbedModelAdmin
from .forms import *
from .inlines import *
from ..models import Contender, SeatsSeason
from ..models import ContenderSeason
from ..models import Driver, Competition, Circuit, Season, Result


def lr_diff(l, r):
    return list(set(l).difference(r))


def lr_intr(l, r):
    return list(set(l).intersection(r))


class DriverAdmin(RelatedCompetitionAdmin, CommonTabbedModelAdmin):
    list_display = ('__unicode__', 'country', 'print_competitions')
    list_filter = ('competitions__name',)
    tab_overview = (
        (None, {
        'fields': ('last_name', 'first_name', 'year_of_birth', 'country')
        }),
    )
    tab_competitions = (
        ContenderInline,
    )
    tabs = [
        ('Overview', tab_overview),
        ('Competitions', tab_competitions)
    ]


class TeamAdmin(RelatedCompetitionAdmin, CommonTabbedModelAdmin):
    model = Team
    list_display = ('__unicode__', 'country', 'print_competitions')
    list_filter = ('competitions', 'seats__seasons')
    tab_overview = (
        (None, {
                'fields': ('name', 'full_name', 'country')
        }),
    )
    tab_competitions = (
        (None, {
            'fields': ('competitions',)
        }),
    )
    tab_seats = (
        SeatInline,
    )
    tabs = [
        ('Overview', tab_overview),
        ('Competitions', tab_competitions),
        ('Seats', tab_seats),
    ]


class CompetitionSeasonAdmin(admin.TabularInline):
    model = Season


class CompetitionAdmin(CommonTabbedModelAdmin):
    model = Competition
    list_display = ('name', 'full_name')
    tab_overview = (
        (None, {
        'fields': ('name', 'full_name', 'country', 'slug')
        }),
    )
    tab_team = (
        CompetitionTeamInline,
    )
    tab_drivers = (
        ContenderInline,
    )
    tab_seasons = (
        CompetitionSeasonAdmin,
    )
    tabs = [
        ('Overview', tab_overview),
        ('Team', tab_team),
        ('Drivers', tab_drivers),
        ('Seasons', tab_seasons)
    ]


class CircuitAdmin(admin.ModelAdmin):
    list_display = ('name', 'city', 'country', 'opened_in')


class GrandPrixAdmin(RelatedCompetitionAdmin, admin.ModelAdmin):
    list_display = ('name', 'country', 'default_circuit', 'print_competitions')
    list_filter = ('competitions',)


class SeatSeasonAdmin(CommonTabbedModelAdmin):
    tab_overview = (
        (None, {
                'fields': ('year', 'competition', 'rounds', 'punctuation')
        }),
    )
    tab_teams = (
        TeamSeasonInline,
    )
    tab_drivers = (
        SeatSeasonInline,
    )
    tabs = [
        ('Overview', tab_overview),
        ('Drivers', tab_drivers),
    ]

    readonly_fields = ('year', 'competition', 'rounds', 'punctuation')
    list_filter = ('competition',)
    list_display = ('competition', 'year',)

    def has_add_permission(self, request):
        return False


class SeasonAdmin(CommonTabbedModelAdmin):
    form = SeasonAdminForm
    tab_overview = (
        (None, {
        'fields': ('year', 'competition', 'rounds', 'punctuation')
        }),
    )
    tab_teams = (
        TeamSeasonInline,
    )
    # tab_drivers = (
    #     SeatSeasonInline,
    # )
    tab_races = (
        RaceInline,
    )
    tabs = [
        ('Overview', tab_overview),
        ('Teams', tab_teams),
        # ('Drivers', tab_drivers),
        ('Races', tab_races),
    ]
    readonly_fields = ('print_copy_season',)
    list_display = ('__unicode__', 'print_copy_season')
    list_filter = ('competition',)

    def get_season_copy(self, copy_id):
        seasons = Season.objects.filter(pk=copy_id)
        copy_dict = {}
        if seasons.count():
            season = seasons.first()
            copy_dict = {
                #'year': season.year+1,
                'competition': season.competition,
                'rounds': season.rounds,
                'punctuation': season.punctuation
            }
        return copy_dict

    def get_changeform_initial_data(self, request, *args, **kwargs):
        copy_id = request.GET.get('copy')
        if copy_id:
            return self.get_season_copy(copy_id)
 
    def print_copy_season(self, obj):
        if obj.pk:
            copy_link = reverse("admin:driver27_season_add")
            get_copy_id = '?copy={obj_pk}'.format(obj_pk=obj.pk)
            return "<a href='{link}{query}'>{copy_text}</a>".format(link=copy_link,
                                                                    query=get_copy_id,
                                                                    copy_text=_('Copy'))
        else:
            return ''
    print_copy_season.short_description = _('copy season')
    print_copy_season.allow_tags = True


class RaceAdmin(CommonTabbedModelAdmin):
    list_display = ('__unicode__', 'season', 'print_pole', 'print_winner', 'print_fastest',)
    list_filter = ('season', 'season__competition',)

    tab_overview = (
        (None, {
            'fields': ('season', 'round', 'grand_prix', 'circuit', 'date', 'alter_punctuation')
        }),
    )
    tab_results = (
        ResultInline,
    )
    tabs = [
        ('Overview', tab_overview),
        # ('Summary', tab_summary),
        # ('Drivers', tab_drivers),
        ('Results', tab_results),
    ]

    # def get_urls(self):
    #     urls = super(RaceAdmin, self).get_urls()
    #     urlpatterns = [
    #         url(r'(?P<race_id>\d+)/results/$', self.admin_site.admin_view(self.results), name='driver27_race_results')
    #     ]
    #
    #     return urlpatterns + urls

    def print_seat(self, seat):
        return u"{driver}".format(driver=seat.contender.driver) if seat else None

    def print_pole(self, obj):
        return self.print_seat(obj.pole)
    print_pole.short_description = _('Pole')

    def print_winner(self, obj):
        return self.print_seat(obj.winner)
    print_winner.short_description = _('Winner')

    def print_fastest(self, obj):
        return self.print_seat(obj.fastest)
    print_fastest.short_description = _('Fastest')

    @staticmethod
    def clean_position(position):
        try:
            position = int(position)
        except (ValueError, TypeError):
            position = None
        return position

    # def edit_result_entries(self, request, entries, race, action='update'):
    #     for entry in entries:
    #         field_prefix = '%s-%s-' % ('seat', str(entry))
    #         qualifying = request.POST.get(field_prefix + 'qualifying', None)
    #         finish = request.POST.get(field_prefix + 'finish', None)
    #         fastest_lap = request.POST.get(field_prefix + 'fastest-lap', False)
    #         retired = request.POST.get(field_prefix + 'retired', False)
    #         wildcard = request.POST.get(field_prefix + 'wildcard', False)
    #
    #         qualifying = self.clean_position(qualifying)
    #         finish = self.clean_position(finish)
    #
    #         dict_to_save = {
    #             'qualifying': qualifying,
    #             'finish': finish,
    #             'fastest_lap': fastest_lap,
    #             'retired': retired,
    #             'wildcard': wildcard
    #         }
    #         if action == 'update':
    #             # filter allows update
    #             result = Result.objects.filter(race=race, seat_id=entry)
    #             result.update(**dict_to_save)
    #         elif action == 'add':
    #             Result.objects.create(
    #                 race=race,
    #                 seat_id=entry,
    #                 **dict_to_save
    #             )
    #
    # def add_result_entries(self, request, entries, race):
    #     self.edit_result_entries(request, entries, race, action='add')
    #
    # def update_result_entries(self, request, entries, race):
    #     self.edit_result_entries(request, entries, race, action='update')
    #
    # def del_result_entries(self, request, entries, race):
    #     for entry in entries:
    #         result = Result.objects.get(race=race, seat_id=entry)
    #         result.delete()
    #
    # def update_race_seats(self, request, new_seats, race):
    #     old_seats = [result.seat_id for result in race.results.all()]
    #     entries_to_add = lr_diff(new_seats, old_seats)
    #     entries_to_upd = lr_intr(new_seats, old_seats)
    #     entries_to_del = lr_diff(old_seats, new_seats)
    #     if entries_to_add:
    #         self.add_result_entries(request, entries_to_add, race)
    #
    #     if entries_to_upd:
    #         self.update_result_entries(request, entries_to_upd, race)
    #
    #     if entries_to_del:
    #         self.del_result_entries(request, entries_to_del, race)
    #
    # def order_entries(self, entries, seats_len):
    #     if seats_len:
    #         entries = sorted(entries, key=lambda x: (
    #             -x['checked'],
    #             -x['finished'],
    #             x['finish'],
    #             -x['qualified'],
    #             x['qualifying'],
    #             -x['season_points']
    #         ))
    #     else:
    #         entries = sorted(entries, key=lambda x: -x['season_points'])
    #     return entries
    #
    # def results(self, request, race_id):
    #     race = self.model.objects.get(pk=race_id)
    #     title = 'Results in %s' % race
    #     season = race.season
    #     if request.method == 'POST':
    #         post_entries = request.POST.getlist('entry[]')
    #         if len(post_entries):
    #             post_entries = list(map(int, post_entries))
    #             self.update_race_seats(request, post_entries, race)
    #
    #         else:
    #             race_seats = [result.seat_id for result in race.results.all()]
    #             self.del_result_entries(request, race_seats, race)
    #     entries = self.get_entries(race, season)
    #
    #     context = {
    #         'race': race, 'season': season, 'entries': entries, 'title': title,
    #         'opts': self.model._meta,
    #         'app_label': self.model._meta.app_label,
    #         'change': True
    #     }
    #     tpl = 'driver27/admin/results.html'
    #     return render(request, tpl, context)
    #
    # def get_entries(self, race, season):
    #     season_seats = Seat.objects.filter(seasons__pk=season.pk)
    #     race_seats = [result.seat_id for result in race.results.all()]
    #     entries = []
    #     for seat in season_seats:
    #         contender = seat.contender
    #         driver_name = ' '.join((contender.driver.first_name, contender.driver.last_name))
    #         season_points = ContenderSeason(contender, season).get_points(limit_races=race.round)
    #
    #         points = finish = qualifying = None
    #         fastest_lap = retired = wildcard = False
    #
    #         is_race_seat = False
    #         if seat.pk in race_seats:
    #             result = Result.objects.get(race=race, seat=seat)
    #             qualifying = result.qualifying
    #             finish = result.finish if result.finish else None
    #             points = result.points
    #             fastest_lap = result.fastest_lap
    #             retired = result.retired
    #             wildcard = result.wildcard
    #             is_race_seat = True
    #
    #         entry = {
    #             'seat': seat.pk,
    #             'driver_name': driver_name,
    #             'team': seat.team.name,
    #             'checked': is_race_seat,
    #             'qualifying': qualifying,
    #             'qualified': bool(qualifying),
    #             'finish': finish,
    #             'finished': bool(finish),
    #             'fastest_lap': fastest_lap,
    #             'retired': retired,
    #             'wildcard': wildcard,
    #             'points': points,
    #             'season_points': season_points
    #         }
    #         entries.append(entry)
    #     entries = self.order_entries(entries, len(race_seats))
    #     return entries


class ContenderAdmin(CommonTabbedModelAdmin):
    list_display = ('__unicode__', 'competition', 'teams_verbose', 'print_current')
    list_filter = ('competition', 'seats__seasons',)
    tab_overview = (
        (None, {
        'fields': ('driver', 'competition')
        }),
    )
    tab_teams = (
        SeatInline,
    )
    tabs = [
        ('Overview', tab_overview),
        ('Teams', tab_teams)
    ]

    def print_current(self, obj):
        filter_current = Seat.objects.filter(
            contender__driver=obj.driver,
            contender__competition=obj.competition,
            current=True)
        return filter_current[0].team if filter_current.count() else None
    print_current.short_description = _('current team')


class SeatAdmin(CommonTabbedModelAdmin):
    list_display = ('contender', 'team',)
    list_filter = ('contender__competition', 'seasons',)
    tab_overview = (
        (None, {
                'fields': ('team', 'contender', 'current')
        }),
    )
    tab_seasons = (
        # (None, {
        #         'fields': ('seasons',)
        # }),
        SeatSeasonInline,
    )
    tabs = [
        ('Overview', tab_overview),
        ('Seasons', tab_seasons)
    ]


admin.site.register(Driver, DriverAdmin)
admin.site.register(Team, TeamAdmin)
admin.site.register(Competition, CompetitionAdmin)
admin.site.register(Circuit, CircuitAdmin)
admin.site.register(GrandPrix, GrandPrixAdmin)
admin.site.register(Season, SeasonAdmin)
admin.site.register(Race, RaceAdmin)
admin.site.register(Seat, SeatAdmin)

# m2m admin
admin.site.register(Contender, ContenderAdmin)
admin.site.register(SeatsSeason, SeatSeasonAdmin)



from django import forms
from django.core.urlresolvers import reverse
from django.utils.translation import ugettext as _
from tabbed_admin import TabbedModelAdmin


# http://stackoverflow.com/a/34567383
class AlwaysChangedModelForm(forms.ModelForm):
    def is_empty_form(self, *args, **kwargs):
        empty_form = True
        for name, field in self.fields.items():
            prefixed_name = self.add_prefix(name)
            data_value = field.widget.value_from_datadict(self.data, self.files, prefixed_name)
            if data_value:
                empty_form = False
                break
        return empty_form

    def has_changed(self, *args, **kwargs):
        """ Should returns True if data differs from initial.
        By always returning true even unchanged inlines will get validated and saved."""
        if self.instance.pk is None and self.initial:
            if not self.changed_data:
                return True
            if self.is_empty_form():
                return False
        return super(AlwaysChangedModelForm, self).has_changed()


class RelatedCompetitionAdmin(object):
    """ Aux class to share print_competitions method between driver and team """
    def print_competitions(self, obj):
        if hasattr(obj, 'competitions'):
            return ', '.join('{competition}'.format(competition=competition)
                             for competition in obj.competitions.all())
        else:
            return None
    print_competitions.short_description = _('competitions')


class CommonTabbedModelAdmin(TabbedModelAdmin):
    def get_form(self, request, obj=None, **kwargs):
        # just save obj reference for future processing in Inline
        if request and obj:
            request._obj_ = obj
        return super(CommonTabbedModelAdmin, self).get_form(request=request, obj=obj, **kwargs)



from .config import get_config


def get_punctuation_config(punctuation_code=None):
    config = get_config('PUNCTUATION', key=punctuation_code)
    return config


def get_punctuation_label_dict():
    config = get_punctuation_config()
    if config:
        return [(punctuation_dict.get('label'), index) for index, punctuation_dict in config.items()]
    return None



# -*- coding: utf-8 -*-
from django.test import TestCase
from django.core.exceptions import ValidationError
from .common import retro_encode, CommonDriverTestCase
from ..models import Driver

class DriverTestCase(TestCase, CommonDriverTestCase):
    def test_driver_unicode(self):
        driver = self.get_test_driver_a()
        expected_unicode = ', '.join((driver.last_name, driver.first_name))
        self.assertEquals(str(driver), retro_encode(expected_unicode))

    def test_driver_save_exception(self):
        driver = self.get_test_driver_a()
        driver.year_of_birth = 1885
        self.assertRaises(ValidationError, driver.save)


import os
from .settings import *

# DATABASES = {
#     'default': {
#         'ENGINE': 'django.db.backends.sqlite3',
#         'NAME': os.path.join(BASE_DIR, 'pytest.sqlite3'),
#     }
# }

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': 'driver27',
        'USER': 'postgres',
#        'PASSWORD': '',
#        'PASSWORD': 'mypassword',
        'HOST': '127.0.0.1',
#        'PORT': '5432',
    }
}

LANGUAGE_CODE = 'es'



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-11-03 08:46
from __future__ import unicode_literals

from django.db import migrations
import exclusivebooleanfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('driver27', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='result',
            name='fastest_lap',
            field=exclusivebooleanfield.fields.ExclusiveBooleanField(default=False, on=('race',)),
        ),
        migrations.AlterField(
            model_name='seat',
            name='current',
            field=exclusivebooleanfield.fields.ExclusiveBooleanField(default=False, on=('contender',)),
        ),
    ]



from django.db.models import Count
from django.http import Http404
from django.shortcuts import render
from django.utils.translation import ugettext as _

from .models import Competition, Season, Race
from .records import get_record_config, get_record_label_dict
from .punctuation import get_punctuation_label_dict


def get_season(slug, year):
    try:
        season = Season.objects.get(year=year, competition__slug=slug)
    except Season.DoesNotExist:
        raise Http404(_('Season does not exist'))
    return season


def competition_view(request, competition_slug=None):
    if competition_slug is not None:
        try:
            competition_obj = Competition.objects.get(slug=competition_slug)
        except Competition.DoesNotExist:
            raise Http404(_('Competition does not exist'))

        title = '{competition}'.format(competition=competition_obj)
        context = {'competition': competition_obj, 'title': title}
        tpl = 'driver27/competition/competition-view.html'
    else:
        competitions = Competition.objects.all()
        title = _('Competitions')
        context = {'competitions': competitions, 'title': title}
        tpl = 'driver27/competition/competition-list.html'
    return render(request, tpl, context)


def season_view(request, competition_slug, year):
    season = get_season(competition_slug, year)
    title = '{competition}/{year:d}'.format(competition=season.competition, year=season.year)
    driver_rank = season.points_rank()[:5]
    team_rank = season.team_points_rank()[:5]
    context = {'season': season, 'title': title, 'driver_rank': driver_rank, 'team_rank': team_rank}
    tpl = 'driver27/season/season-view.html'
    return render(request, tpl, context)


def _rank_view(request, competition_slug, year, rank_model='driver'):
    season = get_season(competition_slug, year)
    scoring_code = request.POST.get('scoring', season.punctuation)
    has_champion = access_to_road = False
    scoring = get_punctuation_label_dict()
    if rank_model == 'driver':
        rank = season.points_rank(scoring_code=scoring_code)
        rank_title = _('DRIVERS')
        tpl = 'driver27/driver/driver-list.html'
        has_champion = season.has_champion()
        access_to_road = (not has_champion and season.pending_races() <= 5)
    elif rank_model == 'team':
        rank = season.team_points_rank()
        rank_title = _('TEAMS')
        tpl = 'driver27/team/team-list.html'
    else:
        raise Http404(_('Impossible rank'))

    title = '{season} [{title}]'.format(season=season, title=rank_title)

    context = {'rank': rank,
               'season': season,
               'title': title,
               'access_to_road': access_to_road,
               'has_champion': has_champion,
               'scoring_list': scoring,
               'scoring_code': scoring_code}
    return render(request, tpl, context)


def driver_rank_view(request, competition_slug, year):
    return _rank_view(request, competition_slug, year, rank_model='driver')


def driver_olympic_view(request, competition_slug, year):
    season = get_season(competition_slug, year)
    rank = season.olympic_rank()
    rank_title = _('DRIVERS rank by olympic mode')
    tpl = 'driver27/driver/driver-list.html'

    title = '{season} [{title}]'.format(season=season, title=rank_title)

    context = {'rank': rank,
               'season': season,
               'title': title,
               'positions': range(1, 21),
               'olympic': True}
    return render(request, tpl, context)


def driver_road_view(request, competition_slug, year):
    season = get_season(competition_slug, year)
    if not season.has_champion():
        pending_races = season.pending_races()
        pending_points = season.pending_points()
        rank = season.points_rank()
        minimum_points_to_road = season.leader[0] - pending_points
        road_rank = [position for position in rank if minimum_points_to_road <= position[0]]
        scoring = season.get_scoring()
        title = _('%(season)s - Road to the championship') % {'season': season}
        tpl = 'driver27/driver/driver-list.html'

        context = {'rank': road_rank,
                   'season': season,
                   'title': title,
                   'scoring': scoring['finish'],
                   'pending_races': range(1, pending_races+1),
                   'road_to_championship': True}
    else:
        raise Http404(_('%(leader)s is the Champion!') % {'leader': season.leader[1]})
    return render(request, tpl, context)


def team_rank_view(request, competition_slug, year):
    return _rank_view(request, competition_slug, year, rank_model='team')


def race_list(request, competition_slug, year):
    season = get_season(competition_slug, year)
    races = season.races.all().order_by('round')
    title = _('%(season)s [RACES]') % {'season': season}
    context = {'races': races, 'season': season, 'title': title}
    tpl = 'driver27/race/race-list.html'
    return render(request, tpl, context)


def race_view(request, competition_slug, year, race_id=None):
    season = get_season(competition_slug, year)
    try:
        race = season.races.get(round=race_id)
    except Race.DoesNotExist:
        raise Http404(_('Race does not exist'))
    results = race.results.all()\
        .annotate(null_position=Count('finish')).order_by('-null_position', 'finish', 'qualifying')
    title = _('Results of %(race)s') % {'race': race}
    context = {'race': race, 'season': season, 'title': title, 'results': results}
    tpl = 'driver27/race/race-view.html'
    return render(request, tpl, context)


def get_safe_record_config(record):
    record_config = get_record_config(record)
    if record_config:
        return record_config
    else:
        raise Http404(_('Record does not exist'))


def get_record_common_context(request, competition_slug, year, record=None):
    season = get_season(competition_slug, year)
    context = {
        'season': season,
        'record': record,
    }
    if record:
        record_config = get_safe_record_config(record)
        title = _('%(record_label)s Record, %(season)s') \
            % {'record_label': record_config.get('label'), 'season': season}
        context['record_filter'] = record_config.get('filter')
    else:
        title = _('Select a %(season)s record') % {'season': season}

    context['title'] = title
    context['record_codes'] = get_record_label_dict()
    return context


def driver_record_view(request, competition_slug, year, record=None):
    context = get_record_common_context(request, competition_slug, year, record)

    rank = None
    if record:
        season = context.get('season')
        rank = season.stats_rank(**context.get('record_filter')) if 'record_filter' in context else None
    context.pop('record_filter', None)
    context['rank'] = rank
    tpl = 'driver27/driver/driver-record.html'
    return render(request, tpl, context)


def team_record_doubles_view(request, competition_slug, year, record=None):
    return team_record_view(request, competition_slug, year, record=record, rank_type='DOUBLES')


def team_record_races_view(request, competition_slug, year, record=None):
    return team_record_view(request, competition_slug, year, record=record, rank_type='RACES')


def team_record_stats_view(request, competition_slug, year, record=None):
    return team_record_view(request, competition_slug, year, record=record, rank_type='STATS')


def team_record_view(request, competition_slug, year, rank_type, record=None):
    context = get_record_common_context(request, competition_slug, year, record)

    rank = None
    if record:
        season = context.get('season')
        rank = season.get_team_rank(rank_type, **context.get('record_filter')) if 'record_filter' in context else None
    context['rank'] = rank
    context['rank_type'] = rank_type

    context['doubles_record_codes'] = get_record_label_dict(doubles=True)
    # context['races'] = (rank_type in ('RACES', 'RACES-DOUBLES'))
    tpl = 'driver27/team/team-record.html'
    return render(request, tpl, context)




# -*- coding: utf-8 -*-
from django.test import TestCase
from .common import CommonSeatTestCase
from ..models import Seat

class SeatTestCase(TestCase, CommonSeatTestCase):
    def test_seat_unicode(self):
        seat = self.get_test_seat()
        contender = seat.contender
        team = seat.team
        expected_seat = '{driver} in {team}/{competition}'.format(driver=str(contender.driver), team=str(team),
                                                                  competition=str(contender.competition))
        self.assertEquals(str(seat), expected_seat)

    def test_seat_current_unique(self):
        seat_a = self.get_test_seat()
        seat_a.current = True
        self.assertIsNone(seat_a.save())
        contender = seat_a.contender
        team_a = seat_a.team
        self.assertEquals(Seat.objects.get(contender=contender, team=team_a).current, True)
        # create seat_c (same contender, different team)
        seat_c = self.get_test_seat_c(seat_a=seat_a)
        seat_c.current = True
        self.assertIsNone(seat_c.save())
        team_c = seat_c.team
        # # Although both seats have current=True, only the LAST is saved.
        self.assertEquals(Seat.objects.filter(contender=contender).count(), 2)
        self.assertEquals(Seat.objects.filter(contender=contender, current=True).count(), 1)
        self.assertEquals(Seat.objects.get(contender=contender, team=team_a).current, False)
        self.assertEquals(Seat.objects.get(contender=contender, team=team_c).current, True)



# -*- coding: utf-8 -*-
from django.test import TestCase
from .common import CommonContenderTestCase

class ContenderTestCase(TestCase, CommonContenderTestCase):
    def test_contender_unicode(self):
        contender = self.get_test_contender()
        expected_str = ' in '.join((str(contender.driver), str(contender.competition)))
        self.assertEquals(str(contender), expected_str)

    def test_contender_team(self):
        contender = self.get_test_contender()
        self.assertEquals(contender.teams_verbose, None)